from typing import Any, Literal, NotRequired, Required, TypedDict

from draive.parameters import DataModel, Field

__all__ = [
    "GeminiGenerationResult",
    "GeminiMessage",
    "GeminiTextMessageContent",
    "GeminiFunctionCallMessageContent",
    "GeminiFunctionResponseMessageContent",
    "GeminiMessageContentBlob",
    "GeminiMessageContentReference",
    "GeminiDataMessageContent",
    "GeminiDataReferenceMessageContent",
    "GeminiFunctionResponse",
    "GeminiFunctionCall",
]


class GeminiMessageContentBlob(DataModel):
    mime_type: str = Field(aliased="mimeType")
    data: str  # base64 encoded


class GeminiMessageContentReference(DataModel):
    mime_type: str = Field(aliased="mimeType")
    uri: str = Field(aliased="fileUri")


class GeminiTextMessageContent(DataModel):
    text: str


class GeminiDataMessageContent(DataModel):
    data: GeminiMessageContentBlob = Field(aliased="inlineData")


class GeminiDataReferenceMessageContent(DataModel):
    reference: GeminiMessageContentReference = Field(aliased="fileData")


class GeminiFunctionCall(DataModel):
    name: str
    arguments: dict[str, Any] = Field(aliased="args")


class GeminiFunctionCallMessageContent(DataModel):
    function_call: GeminiFunctionCall = Field(aliased="functionCall")


class GeminiFunctionResponse(DataModel):
    name: str
    response: dict[str, Any]


class GeminiFunctionResponseMessageContent(DataModel):
    function_response: GeminiFunctionResponse = Field(aliased="functionResponse")


GeminiMessageContent = (
    GeminiTextMessageContent
    | GeminiFunctionCallMessageContent
    | GeminiFunctionResponseMessageContent
    | GeminiDataMessageContent
    | GeminiDataReferenceMessageContent
)


class GeminiMessage(DataModel):
    role: Literal["user", "model"]
    content: list[GeminiMessageContent] = Field(aliased="parts")


class GeminiRequestMessage(TypedDict, total=False):
    role: Required[Literal["user", "model"]]
    parts: Required[list[dict[str, Any]]]


class GeminiSystemMessageContent(TypedDict, total=False):
    text: Required[str]


class GeminiSystemMessage(TypedDict, total=False):
    parts: Required[tuple[GeminiSystemMessageContent]]


class GeminiFunctionToolSpecification(TypedDict, total=False):
    name: Required[str]
    description: Required[str]
    parameters: NotRequired[dict[str, Any]]


class GeminiFunctionsTool(TypedDict, total=False):
    functionDeclarations: Required[list[GeminiFunctionToolSpecification]]


class GeminiUsage(DataModel):
    prompt_tokens: int = Field(aliased="promptTokenCount", default_factory=int)
    generated_tokens: int = Field(aliased="candidatesTokenCount", default_factory=int)


class GeminiChoice(DataModel):
    content: GeminiMessage | None = None
    finish_reason: Literal[
        "STOP",
        "MAX_TOKENS",
        "SAFETY",
        "RECITATION",
        "OTHER",
    ] = Field(aliased="finishReason")


class GeminiGenerationResult(DataModel):
    choices: list[GeminiChoice] = Field(aliased="candidates")
    usage: GeminiUsage = Field(aliased="usageMetadata")



from draive.tokenization.call import count_text_tokens, tokenize_text
from draive.tokenization.state import Tokenization
from draive.tokenization.text import TextTokenizer

__all__ = [
    "count_text_tokens",
    "TextTokenizer",
    "Tokenization",
    "tokenize_text",
]



import json
from http import HTTPStatus
from typing import Any, Literal, Self, cast, final

from httpx import AsyncClient, Response

from draive.ollama.config import OllamaChatConfig
from draive.ollama.errors import OllamaException
from draive.ollama.models import ChatCompletionResponse, ChatMessage
from draive.parameters import DataModel
from draive.scope import ScopeDependency
from draive.utils import getenv_str, not_missing

__all__ = [
    "OllamaClient",
]


@final
class OllamaClient(ScopeDependency):
    @classmethod
    def prepare(cls) -> Self:
        return cls(
            endpoint=getenv_str("OLLAMA_ENDPOINT", "http://localhost:11434"),
            timeout=90,
        )

    def __init__(
        self,
        endpoint: str,
        timeout: float | None = None,
    ) -> None:
        self._client: AsyncClient = AsyncClient(
            base_url=endpoint,
            timeout=timeout,
        )

    async def chat_completion(
        self,
        *,
        config: OllamaChatConfig,
        messages: list[ChatMessage],
    ) -> ChatCompletionResponse:
        return await self._create_chat_completion(
            messages=messages,
            model=config.model,
            temperature=config.temperature,
            top_k=config.top_k if not_missing(config.top_k) else None,
            top_p=config.top_p if not_missing(config.top_p) else None,
            max_tokens=config.max_tokens if not_missing(config.max_tokens) else None,
            response_format=cast(Literal["text", "json"], config.response_format)
            if not_missing(config.response_format)
            else "text",
            seed=config.seed if not_missing(config.seed) else None,
        )

    async def _create_chat_completion(  # noqa: PLR0913
        self,
        model: str,
        temperature: float,
        top_k: float | None,
        top_p: float | None,
        seed: int | None,
        max_tokens: int | None,
        response_format: Literal["text", "json"],
        messages: list[ChatMessage],
    ) -> ChatCompletionResponse:
        request_body: dict[str, Any] = {
            "model": model,
            "messages": [message.as_dict() for message in messages],
            "options": {
                "temperature": temperature,
            },
            "stream": False,
        }

        if max_tokens:
            request_body["options"]["num_predict"] = max_tokens
        if top_k is not None:
            request_body["options"]["top_k"] = top_k
        if top_p is not None:
            request_body["options"]["top_p"] = top_p
        if seed is not None:
            request_body["options"]["seed"] = seed
        if response_format == "json":
            request_body["format"] = "json"

        return await self._request(
            model=ChatCompletionResponse,
            method="POST",
            url="/api/chat",
            body=request_body,
        )

    async def dispose(self) -> None:
        await self._client.aclose()

    async def _request[Requested: DataModel](  # noqa: PLR0913
        self,
        model: type[Requested],
        method: str,
        url: str,
        query: dict[str, str] | None = None,
        headers: dict[str, str] | None = None,
        body: DataModel | dict[str, Any] | None = None,
        follow_redirects: bool | None = None,
        timeout: float | None = None,
    ) -> Requested:
        request_headers: dict[str, str]
        if headers:
            request_headers = headers
        else:
            request_headers = {
                "Accept": "application/json",
            }

        body_content: str | None
        match body:
            case None:
                body_content = None

            case body_model if isinstance(body_model, DataModel):
                body_content = body_model.as_json()

            case values:
                body_content = json.dumps(values)

        if body_content:
            request_headers["Content-Type"] = "application/json"

        response: Response
        try:
            response = await self._client.request(
                method=method,
                url=url,
                headers=request_headers,
                params=query,
                content=body_content,
                follow_redirects=follow_redirects or False,
                timeout=timeout,
            )

        except Exception as exc:
            raise OllamaException("Network request failed") from exc

        status: HTTPStatus = HTTPStatus(value=response.status_code)
        if status.is_success:
            try:
                return model.from_json(await response.aread())

            except Exception as exc:
                raise OllamaException("Failed to decode Ollama response", response) from exc

        elif status.is_client_error:
            error_body: bytes = await response.aread()
            raise OllamaException(
                "Ollama request error: %s %s",
                status,
                error_body.decode("utf-8"),
            )

        else:
            raise OllamaException("Network request failed", response)



from draive.scope.access import ctx
from draive.scope.dependencies import ScopeDependencies, ScopeDependency
from draive.scope.state import ScopeState

__all__ = [
    "ctx",
    "ScopeDependencies",
    "ScopeDependency",
    "ScopeState",
]



from draive.lmm.call import lmm_invocation
from draive.lmm.invocation import LMMInvocation, LMMToolSelection
from draive.lmm.state import LMM
from draive.lmm.tools import (
    AnyTool,
    Tool,
    ToolAvailabilityCheck,
    Toolbox,
    ToolContext,
    ToolException,
    ToolSpecification,
    ToolStatus,
    tool,
)

__all__ = [
    "AnyTool",
    "lmm_invocation",
    "LMM",
    "LMMInvocation",
    "LMMToolSelection",
    "tool",
    "Tool",
    "ToolAvailabilityCheck",
    "Toolbox",
    "ToolContext",
    "ToolException",
    "ToolSpecification",
    "ToolStatus",
]



import json
from asyncio import gather
from collections.abc import Sequence
from http import HTTPStatus
from itertools import chain
from typing import Any, Literal, Self, final, overload

from httpx import AsyncClient, Response

from draive.gemini.config import GeminiConfig, GeminiEmbeddingConfig
from draive.gemini.errors import GeminiException
from draive.gemini.models import (
    GeminiFunctionsTool,
    GeminiGenerationResult,
    GeminiRequestMessage,
)
from draive.parameters import DataModel
from draive.scope import ScopeDependency
from draive.utils import getenv_str, not_missing

__all__ = [
    "GeminiClient",
]


@final
class GeminiClient(ScopeDependency):
    @classmethod
    def prepare(cls) -> Self:
        return cls(
            endpoint=getenv_str("GEMINI_ENDPOINT", "https://generativelanguage.googleapis.com"),
            api_key=getenv_str("GEMINI_API_KEY"),
            timeout=90,
        )

    def __init__(
        self,
        endpoint: str,
        api_key: str | None,
        timeout: float | None = None,
    ) -> None:
        self._client: AsyncClient = AsyncClient(
            base_url=endpoint,
            params={
                "key": api_key,
            },
            timeout=timeout,
        )

    async def generate(  # noqa: PLR0913
        self,
        *,
        config: GeminiConfig,
        instruction: str,
        messages: list[GeminiRequestMessage],
        tools: list[GeminiFunctionsTool] | None = None,
        tool_calling_mode: Literal["AUTO", "ANY", "NONE"] = "AUTO",
        response_schema: dict[str, Any] | None = None,
        stream: bool = False,
    ) -> GeminiGenerationResult:
        if stream:
            raise NotImplementedError("Gemini streaming is not supported yet")

        else:
            return await self._generate_content(
                model=config.model,
                request={
                    "generationConfig": {
                        "responseMimeType": config.response_format
                        if not_missing(config.response_format)
                        else "text/plain",
                        "temperature": config.temperature,
                        "topP": config.top_p if not_missing(config.top_p) else None,
                        "topK": config.top_k if not_missing(config.top_k) else None,
                        "maxOutputTokens": config.max_tokens,
                        "responseSchema": response_schema if response_schema else None,
                        "candidateCount": 1,
                    },
                    "systemInstruction": {
                        "parts": ({"text": instruction},),
                    },
                    "contents": messages,
                    "tools": tools or [],
                    "toolConfig": {
                        "functionCallingConfig": {
                            "mode": tool_calling_mode if tools else "NONE",
                        },
                    },
                    "safetySettings": [  # google moderation is terrible, disabling it all
                        {
                            "category": "HARM_CATEGORY_HATE_SPEECH",
                            "threshold": "BLOCK_NONE",
                        },
                        {
                            "category": "HARM_CATEGORY_SEXUALLY_EXPLICIT",
                            "threshold": "BLOCK_NONE",
                        },
                        {
                            "category": "HARM_CATEGORY_DANGEROUS_CONTENT",
                            "threshold": "BLOCK_NONE",
                        },
                        {
                            "category": "HARM_CATEGORY_HARASSMENT",
                            "threshold": "BLOCK_NONE",
                        },
                    ],
                },
            )

    async def _generate_content(
        self,
        model: str,
        request: dict[str, Any],
    ) -> GeminiGenerationResult:
        return await self._request(
            model=GeminiGenerationResult,
            method="POST",
            url=f"v1beta/models/{model}:generateContent",
            body=request,
        )

    async def embedding(
        self,
        config: GeminiEmbeddingConfig,
        inputs: Sequence[str],
    ) -> list[list[float]]:
        return list(
            chain(
                *await gather(
                    *[
                        self._create_text_embedding(
                            model=config.model,
                            texts=inputs[index : index + config.batch_size],
                        )
                        for index in range(0, len(inputs), config.batch_size)
                    ]
                )
            )
        )

    async def _create_text_embedding(
        self,
        model: str,
        texts: Sequence[str],
    ) -> list[list[float]]:
        if not texts:
            return []

        response: dict[str, Any] = await self._request(
            model=None,
            method="POST",
            url=f"v1beta/models/{model}:batchEmbedText",
            body={"texts": texts},
        )

        result: list[list[float]] = []
        match response:
            case {"embeddings": [*embeddings]}:
                for embedding in embeddings:
                    match embedding:
                        case {"value": [*vector]}:
                            assert all(isinstance(value, float) for value in vector)  # nosec: B101
                            result.append(vector)

                        case _:
                            raise GeminiException("Invalid Gemini embedding response: %s", response)

            case _:
                raise GeminiException("Invalid Gemini embedding response: %s", response)

        assert len(texts) == len(result)  # nosec: B101
        return result

    async def dispose(self) -> None:
        await self._client.aclose()

    @overload
    async def _request[Requested: DataModel](
        self,
        model: None,
        method: str,
        url: str,
        query: dict[str, str] | None = None,
        headers: dict[str, str] | None = None,
        body: DataModel | dict[str, Any] | None = None,
        follow_redirects: bool | None = None,
        timeout: float | None = None,
    ) -> dict[str, Any]: ...

    @overload
    async def _request[Requested: DataModel](
        self,
        model: type[Requested],
        method: str,
        url: str,
        query: dict[str, str] | None = None,
        headers: dict[str, str] | None = None,
        body: DataModel | dict[str, Any] | None = None,
        follow_redirects: bool | None = None,
        timeout: float | None = None,
    ) -> Requested: ...

    async def _request[Requested: DataModel](  # noqa: PLR0913, PLR0912
        self,
        model: type[Requested] | None,
        method: str,
        url: str,
        query: dict[str, str] | None = None,
        headers: dict[str, str] | None = None,
        body: DataModel | dict[str, Any] | None = None,
        follow_redirects: bool | None = None,
        timeout: float | None = None,
    ) -> Requested | dict[str, Any]:
        request_headers: dict[str, str]
        if headers:
            request_headers = headers

        else:
            request_headers = {
                "Accept": "application/json",
            }

        body_content: str | None
        match body:
            case None:
                body_content = None

            case body_model if isinstance(body_model, DataModel):
                body_content = body_model.as_json()

            case values:
                body_content = json.dumps(values)

        if body_content:
            request_headers["Content-Type"] = "application/json"

        response: Response
        try:
            response = await self._client.request(
                method=method,
                url=url,
                headers=request_headers,
                params=query,
                content=body_content,
                follow_redirects=follow_redirects or False,
                timeout=timeout,
            )

        except Exception as exc:
            raise GeminiException("Network request failed") from exc

        status: HTTPStatus = HTTPStatus(value=response.status_code)
        if status.is_success:
            try:
                if model := model:
                    return model.from_json(await response.aread())

                else:
                    return json.loads(await response.aread())

            except Exception as exc:
                raise GeminiException("Failed to decode Gemini response", response) from exc

        elif status.is_client_error:
            error_body: bytes = await response.aread()
            raise GeminiException(
                "Gemini request error: %s, %s",
                status,
                error_body.decode("utf-8"),
            )

        else:
            raise GeminiException("Network request failed: %s", response)



from draive.ollama.client import OllamaClient
from draive.ollama.config import OllamaChatConfig
from draive.ollama.errors import OllamaException
from draive.ollama.lmm import ollama_lmm_invocation

__all__ = [
    "ollama_lmm_invocation",
    "OllamaChatConfig",
    "OllamaClient",
    "OllamaException",
]



from typing import Any

from draive.scope import ctx
from draive.tokenization.state import Tokenization

__all__ = [
    "tokenize_text",
    "count_text_tokens",
]


def tokenize_text(
    text: str,
    **extra: Any,
) -> list[int]:
    return ctx.state(Tokenization).tokenize_text(
        text=text,
        **extra,
    )


def count_text_tokens(
    text: str,
    **extra: Any,
) -> int:
    return len(
        ctx.state(Tokenization).tokenize_text(
            text=text,
            **extra,
        )
    )



from collections.abc import Generator

__all__ = [
    "markdown_block",
    "markdown_blocks",
]


def markdown_blocks(  # noqa: C901, PLR0912
    info: str = "",
    /,
    *,
    source: str,
) -> Generator[str, None]:
    boundary_sequence: str = "```"
    opening_sequence: str = f"{boundary_sequence}{info}"

    accumulator: str = ""
    content: str = ""
    in_sequence: bool = False
    in_content: bool = False

    for char in source:
        if in_sequence:
            if accumulator.startswith(boundary_sequence):
                if char.isspace():
                    if in_content:
                        in_content = False
                        yield content
                        content = ""  # clear to be ready for next block

                    elif accumulator.startswith(opening_sequence):
                        in_content = True
                        content = ""  # clear current content in case of nested blocks

                    in_sequence = False
                    accumulator = ""

                elif char == "`":
                    if in_content:
                        content += char
                    # keep the accumulator unchanged

                elif char.isalnum() and not in_content:  # take the info
                    accumulator += char

                elif in_content:
                    accumulator += char
                    content += accumulator
                    in_sequence = False
                    accumulator = ""

                else:
                    in_sequence = False
                    accumulator = ""

            elif char == "`":
                accumulator += char

            elif in_content:
                in_sequence = False
                accumulator += char
                content += accumulator
                accumulator = ""

            else:
                in_sequence = False
                accumulator = ""

        elif char == "`":
            in_sequence = True
            accumulator = char

        elif in_content:
            content += char

        # else skip character

    # when we hit the end while closing sequence was in place yield the last part
    if in_content and in_sequence and accumulator == boundary_sequence:
        yield content


def markdown_block(
    info: str = "",
    /,
    *,
    source: str,
) -> str | None:
    try:
        return next(
            markdown_blocks(
                info,
                source=source,
            )
        )

    except StopIteration:
        return None



from asyncio import AbstractEventLoop, CancelledError, Task, TimerHandle, gather, get_running_loop
from inspect import isfunction
from typing import Protocol, cast, final, overload, runtime_checkable
from uuid import UUID, uuid4

from draive.agents.definition import AgentMessage, AgentNode
from draive.agents.errors import AgentException
from draive.agents.idle import IdleMonitor
from draive.agents.node import Agent, AgentOutput
from draive.agents.runner import AgentRunner
from draive.helpers import VolatileMemory
from draive.parameters import DataModel, ParametrizedData, State
from draive.scope import ctx
from draive.types import (
    BasicMemory,
    Memory,
    MultimodalContent,
    MultimodalContentConvertible,
    frozenlist,
)
from draive.utils import MISSING, AsyncQueue, Missing, freeze, not_missing

__all__ = [
    "workflow",
    "AgentWorkflow",
    "AgentWorkflowInvocation",
    "AgentWorkflowOutput",
    "AgentWorkflowIdle",
    "AgentWorkflowInput",
]


type AgentWorkflowInput = AgentMessage | AgentException
type AgentWorkflowOutput[Result: ParametrizedData | str] = (
    frozenlist[AgentMessage] | AgentMessage | Result | None
)


@runtime_checkable
class AgentWorkflowStateInitializer[AgentWorkflowState](Protocol):
    def __call__(self) -> AgentWorkflowState: ...


@runtime_checkable
class AgentWorkflowInvocation[AgentWorkflowState, AgentWorkflowResult: ParametrizedData | str](
    Protocol
):
    async def __call__(
        self,
        memory: BasicMemory[AgentWorkflowState],
        input: AgentWorkflowInput,  # noqa: A002
    ) -> AgentWorkflowOutput[AgentWorkflowResult]: ...


class AgentWorkflowIdle(AgentException):
    pass


@final
class AgentWorkflow[AgentWorkflowState, AgentWorkflowResult: ParametrizedData | str]:
    def __init__(
        self,
        node: AgentNode,
        invocation: AgentWorkflowInvocation[
            AgentWorkflowState,
            AgentWorkflowResult,
        ],
        state_initializer: AgentWorkflowStateInitializer[AgentWorkflowState],
    ) -> None:
        self.node: AgentNode = node
        self._invocation: AgentWorkflowInvocation[
            AgentWorkflowState,
            AgentWorkflowResult,
        ] = invocation
        self._state_initializer: AgentWorkflowStateInitializer[AgentWorkflowState] = (
            state_initializer
        )

        freeze(self)

    async def __call__(
        self,
        memory: BasicMemory[AgentWorkflowState],
        input: AgentWorkflowInput,  # noqa: A002
    ) -> AgentWorkflowOutput[AgentWorkflowResult]:
        return await self._invocation(
            memory=memory,
            input=input,
        )

    async def run(
        self,
        input: MultimodalContent | MultimodalContentConvertible,  # noqa: A002
        state: AgentWorkflowState | None = None,
        timeout: float = 120,  # default timeout is 2 minutes
    ) -> AgentWorkflowResult:
        return await WorkflowRunner[AgentWorkflowState, AgentWorkflowResult].run(
            self,
            input=input,
            memory=VolatileMemory(state or self._state_initializer()),
            timeout=timeout,
        )

    def address(
        self,
        content: MultimodalContent | MultimodalContentConvertible,
        /,
        *_content: MultimodalContent | MultimodalContentConvertible,
        addressee: AgentNode | None = None,
    ) -> AgentMessage:
        return AgentMessage(
            identifier=uuid4(),
            sender=MISSING,
            recipient=self.node,
            addressee=addressee,
            content=MultimodalContent.of(content, *_content),
            responding=None,
        )


class PartialAgentWorkflowWrapper[AgentWorkflowState](Protocol):
    def __call__[AgentWorkflowResult: ParametrizedData | str](
        self,
        invocation: AgentWorkflowInvocation[AgentWorkflowState, AgentWorkflowResult],
    ) -> AgentWorkflow[AgentWorkflowState, AgentWorkflowResult]: ...


@overload
def workflow[AgentWorkflowState](
    node: AgentNode,
    /,
    *,
    state: AgentWorkflowStateInitializer[AgentWorkflowState],
) -> PartialAgentWorkflowWrapper[AgentWorkflowState]: ...


@overload
def workflow[AgentWorkflowState](
    *,
    name: str | None = None,
    description: str,
    state: AgentWorkflowStateInitializer[AgentWorkflowState],
) -> PartialAgentWorkflowWrapper[AgentWorkflowState]: ...


def workflow[AgentWorkflowState](
    node: AgentNode | None = None,
    *,
    name: str | None = None,
    description: str | None = None,
    state: AgentWorkflowStateInitializer[AgentWorkflowState],
) -> PartialAgentWorkflowWrapper[AgentWorkflowState]:
    assert node is None or (  # nosec: B101
        name is None and description is None
    ), "Can't specify both agent node and name/description"
    assert (  # nosec: B101
        description is not None or node is not None
    ), "Either agent node or description has to be provided"

    def wrap[AgentWorkflowResult: ParametrizedData | str](
        invocation: AgentWorkflowInvocation[AgentWorkflowState, AgentWorkflowResult],
    ) -> AgentWorkflow[AgentWorkflowState, AgentWorkflowResult]:
        assert isfunction(invocation), "workflow has to be defined from a function"  # nosec: B101

        workflow_node: AgentNode = node or AgentNode(
            name=name or invocation.__qualname__,
            description=description or "",
        )

        def initialize_agent() -> Agent:
            agent_memory: BasicMemory[AgentWorkflowState] = cast(
                BasicMemory[AgentWorkflowState], VolatileMemory(state())
            )

            async def agent(message: AgentMessage) -> AgentOutput:
                match await cast(  # pyright fails to properly type this function
                    AgentWorkflowInvocation[AgentWorkflowState, AgentWorkflowResult],
                    invocation,
                )(
                    memory=agent_memory,
                    input=message,
                ):
                    case None:
                        pass

                    case tuple() as messages:
                        return messages

                    case AgentMessage() as message:
                        return message

                    case result:
                        return message.respond(
                            MultimodalContent.of(
                                result if isinstance(result, DataModel) else str(result)
                            )
                        )

            return agent

        # workflow can run as an agent with limited capabilities
        workflow_node.associate(
            initialize_agent,
            concurrent=False,  # workflow can't be concurrent
        )

        return AgentWorkflow[AgentWorkflowState, AgentWorkflowResult](
            node=workflow_node,
            invocation=invocation,
            state_initializer=state,
        )

    return wrap


@runtime_checkable
class WorkflowRunnerOutput(Protocol):
    def __call__(
        self,
        messages: AgentMessage,
        /,
        *_messages: AgentMessage,
    ) -> None: ...


@runtime_checkable
class WorkflowRunnerResultOutput[WorkflowResult: ParametrizedData | str](Protocol):
    def __call__(
        self,
        result: WorkflowResult,
        /,
    ) -> None: ...


class WorkflowHistory(State):
    history: frozenlist[AgentMessage]


class WorkflowRunner[WorkflowState, WorkflowResult: ParametrizedData | str]:
    @classmethod
    async def run(
        cls,
        workflow: AgentWorkflow[WorkflowState, WorkflowResult],
        /,
        input: MultimodalContent | MultimodalContentConvertible,  # noqa: A002
        memory: Memory[WorkflowState, WorkflowState],
        timeout: float,
    ) -> WorkflowResult:
        return await cls(
            workflow=workflow,
            memory=memory,
            loop=get_running_loop(),
        ).execute(
            input=input,
            timeout=timeout,
        )

    def __init__(
        self,
        workflow: AgentWorkflow[WorkflowState, WorkflowResult],
        memory: Memory[WorkflowState, WorkflowState],
        loop: AbstractEventLoop,
    ) -> None:
        self._loop: AbstractEventLoop = loop
        self._history: list[AgentMessage] = []
        self._workflow_queue: AsyncQueue[AgentMessage | AgentException] = AsyncQueue()
        self._workflow_memory: Memory[WorkflowState, WorkflowState] = memory
        self._workflow: AgentWorkflow[WorkflowState, WorkflowResult] = workflow
        self._agent_runners: dict[UUID, AgentRunner] = {}
        self._status: IdleMonitor = IdleMonitor()
        self._result: WorkflowResult | BaseException | Missing = MISSING

    def __del__(self) -> None:
        self.finish(result=CancelledError())

    @overload
    def send(
        self,
        error: AgentException,
        /,
    ) -> None: ...

    @overload
    def send(
        self,
        messages: AgentMessage,
        /,
        *_messages: AgentMessage,
    ) -> None: ...

    def send(
        self,
        messages: AgentMessage | AgentException,
        /,
        *_messages: AgentMessage,
    ) -> None:
        pending: list[AgentMessage | AgentException] = [messages, *_messages]
        for _ in pending:  # ensure not idle when received a message
            self._status.enter_task()
        self._history.extend([message for message in pending if isinstance(message, AgentMessage)])
        self._workflow_queue.enqueue(*pending)

    @property
    def finished(self) -> bool:
        return not_missing(self._result)

    def finish(
        self,
        result: WorkflowResult | BaseException,
    ) -> None:
        if not_missing(self._result):
            return  # already finished

        ctx.log_debug("Finishing workflow with result: %s", result)

        self._result = result

        for runner in self._agent_runners.values():
            runner.finish()

        self._workflow_queue.finish()

    async def _wait(self) -> None:
        await gather(
            *[runner.wait() for runner in self._agent_runners.values()],
            return_exceptions=False,
        )

    async def finalize(
        self,
        result: WorkflowResult | BaseException,
    ) -> None:
        self.finish(result=result)
        await self._wait()

    async def execute(  # noqa: PLR0912
        self,
        input: MultimodalContent | MultimodalContentConvertible,  # noqa: A002
        timeout: float,
    ) -> WorkflowResult:
        assert not self._agent_runners, "WorkflowRunner can run only once!"  # nosec: B101

        if self._workflow_queue.finished or not_missing(self._result):
            raise RuntimeError("WorkflowRunner can run only once!")

        async with ctx.nested(self._workflow.node.__str__()):

            def on_timeout() -> None:
                self.finish(result=TimeoutError())

            timeout_handle: TimerHandle = self._loop.call_later(
                delay=timeout,
                callback=on_timeout,
            )

            self.send(self._workflow.address(input))

            idle_monitor_task: Task[None] = ctx.spawn_subtask(self._idle_monitor)

            try:
                async for element in self._workflow_queue:
                    try:
                        match element:
                            case AgentMessage() as message:
                                if message.recipient.identifier == self._workflow.node.identifier:
                                    await self._handle(message)

                                elif runner := self._agent_runners.get(
                                    message.recipient.identifier
                                ):
                                    runner.send(message)

                                else:
                                    spawned_runner: AgentRunner = AgentRunner.run(
                                        message.recipient,
                                        output=self.send,
                                        status=self._status.nested(),
                                    )
                                    self._agent_runners[message.recipient.identifier] = (
                                        spawned_runner
                                    )
                                    spawned_runner.send(message)

                            case error:
                                await self._handle(error)

                    finally:
                        self._status.exit_task()

            except CancelledError:
                pass  # just end on cancel

            finally:  # cleanup status when exiting
                timeout_handle.cancel()  # cancel the timeout
                idle_monitor_task.cancel()  # finish idle monitor
                self._status.exit_all()

            await self._wait()  # wait for completion of all runners

            ctx.record(WorkflowHistory(history=tuple(self._history)))

            match self._result:
                case BaseException() as exc:
                    raise exc

                case Missing():
                    raise RuntimeError("Invalid workflow state")

                case result:
                    return result

    async def _handle(
        self,
        input: AgentWorkflowInput,  # noqa: A002
        /,
    ) -> None:
        try:
            match await self._workflow(
                memory=self._workflow_memory,
                input=input,
            ):
                case None:
                    pass  # nothing to do

                case [*messages]:
                    self.send(
                        *[message.updated(sender=self._workflow.node) for message in messages]
                    )

                case AgentMessage() as message:
                    self.send(message.updated(sender=self._workflow.node))

                case result:
                    self.finish(result=cast(WorkflowResult, result))

        except BaseException as exc:
            self.finish(result=exc)

    async def _idle_monitor(self) -> None:
        while not self.finished:  # run until finished
            await self._status.wait_idle()  # wait until all agents become idle

            if self.finished:  # check if that is not due to finishing
                return  # workflow might be finished at this stage

            ctx.log_warning("Detected workflow idle state, attempting to resume...")

            # when workflow should still be running (did not provided result yet)
            # but no agent is working then notify workflow node about the error
            self.send(AgentWorkflowIdle())



from draive.evaluation import EvaluationScore, evaluator
from draive.evaluators.score import CommonScoreModel
from draive.generation import generate_model

__all__ = [
    "text_consistency_evaluator",
]


INSTRUCTION: str = """\
You will be given a reference text and a compared text based on the reference text.
Your task is to rate the compared text using only the Consistency metric, \
which is described in the Evaluation Criteria.
Please make sure you read and understand these instructions very carefully.
Keep this document open while reviewing, and refer to it as needed.

Evaluation Criteria:
Consistency(0.0-4.0) - the factual alignment between the reference text and the compared text.
A factually consistent compared text contains only statements that are entailed \
by the reference text.
Annotators should penalize compared texts that contain hallucinated facts.

Rating Scale:
0.0: Very low consistency - the text contains multiple hallucinated facts \
or significant misalignments with the reference text.
1.0: Low consistency - the text has several instances of information not supported by \
the reference text.
2.0: Moderate consistency - the text is mostly consistent but contains a few unsupported statements.
3.0: Good consistency - the text is largely consistent with minor discrepancies.
4.0: Excellent consistency - the text is fully consistent with the reference text, \
containing only supported information.

Evaluation Steps:
1. Read the compared text and the reference text carefully.
2. Compare the compared text to the reference text and identify the main points \
of the reference text.
3. Assess how well the compared text covers the main points of the reference text \
and how much irrelevant or redundant information it contains.
4. Assign a consistency score from 0.0 to 4.0 based on the provided criteria.

Important: The score must be a decimal number from 0.0 to 4.0. 4.0 is the maximum, \
do not exceed this value.
"""


INPUT_TEMPLATE: str = """
<REFERENCE_TEXT>
{reference}
</REFERENCE_TEXT>

<COMPARED_TEXT>
{compared}
</COMPARED_TEXT>
"""


@evaluator(name="text_consistency")
async def text_consistency_evaluator(
    compared: str,
    /,
    reference: str,
) -> EvaluationScore:
    if not compared:
        return EvaluationScore(
            value=0,
            comment="Input text was empty!",
        )

    if not reference:
        return EvaluationScore(
            value=0,
            comment="Reference text was empty!",
        )

    score: CommonScoreModel = await generate_model(
        CommonScoreModel,
        instruction=INSTRUCTION,
        input=INPUT_TEMPLATE.format(
            reference=reference,
            compared=compared,
        ),
        examples=[
            (
                INPUT_TEMPLATE.format(
                    reference=(
                        "Dolphins are intelligent marine mammals. They use echolocation "
                        "to navigate and hunt. Dolphins live in social groups called pods."
                    ),
                    compared=(
                        "Dolphins are smart fish that can fly short distances. They use sonar "
                        "to talk to whales. Dolphins live in families and go to school "
                        "to learn hunting techniques."
                    ),
                ),
                CommonScoreModel(score=0.0),
            ),
            (
                INPUT_TEMPLATE.format(
                    reference=(
                        "Coffee is a popular beverage worldwide. "
                        "It's made from roasted coffee beans. Caffeine in coffee "
                        "can boost energy and alertness. However, excessive consumption may "
                        "lead to sleep issues."
                    ),
                    compared=(
                        "Coffee is a widely consumed drink around the world. It's produced "
                        "by roasting coffee beans. The caffeine in coffee can increase energy "
                        "levels and improve alertness. However, drinking too much coffee might "
                        "cause sleep problems. Coffee is also known to improve memory and reduce "
                        "the risk of certain diseases."
                    ),
                ),
                CommonScoreModel(score=2.0),
            ),
            (
                INPUT_TEMPLATE.format(
                    reference=(
                        "Photosynthesis is the process by which plants use sunlight to "
                        "produce energy. It requires water, carbon dioxide, and chlorophyll. "
                        "Oxygen is released as a byproduct of photosynthesis."
                    ),
                    compared=(
                        "Plants carry out photosynthesis to create energy from sunlight. "
                        "This process needs water, carbon dioxide, and the green pigment "
                        "chlorophyll. As plants photosynthesize, "
                        "they release oxygen into the environment."
                    ),
                ),
                CommonScoreModel(score=4.0),
            ),
        ],
    )
    return score.normalized(divider=4)



from typing import Any, Protocol, runtime_checkable

from draive.instructions import Instruction
from draive.types import ImageContent, MultimodalContent, MultimodalContentConvertible

__all__ = [
    "ImageGenerator",
]


@runtime_checkable
class ImageGenerator(Protocol):
    async def __call__(
        self,
        *,
        instruction: Instruction | str,
        input: MultimodalContent | MultimodalContentConvertible | None = None,  # noqa: A002
        **extra: Any,
    ) -> ImageContent: ...



from asyncio import Lock
from collections.abc import AsyncIterable, Generator, Iterator
from concurrent.futures import Executor, ThreadPoolExecutor
from typing import Any, Final, Literal, Self, cast, final, overload

from mistralrs import (  # type: ignore
    Architecture,
    ChatCompletionChunkResponse,
    ChatCompletionRequest,
    ChatCompletionResponse,
    Runner,
    Which,
)

from draive.mrs.config import MRSChatConfig
from draive.mrs.errors import MRSException
from draive.scope import ScopeDependency, ctx
from draive.utils import asynchronous, not_missing

__all__ = [
    "MRSClient",
]

MRS_EXECUTOR: Final[Executor | None] = ThreadPoolExecutor(max_workers=1)


@final
class MRSClient(ScopeDependency):
    @classmethod
    def prepare(cls) -> Self:
        return cls(
            models={
                "Phi-3": Which.Plain(
                    model_id="microsoft/Phi-3-mini-4k-instruct",
                    arch=Architecture.Phi3,
                    tokenizer_json=None,
                    repeat_last_n=64,
                ),
            }
        )

    def __init__(
        self,
        models: dict[
            str,
            Which.Plain
            | Which.Lora
            | Which.XLora
            | Which.GGUF
            | Which.GGML
            | Which.LoraGGML
            | Which.LoraGGUF
            | Which.XLoraGGML
            | Which.XLoraGGUF
            | Which.VisionPlain,
        ],
    ) -> None:
        self._models: dict[
            str,
            Which.Plain
            | Which.Lora
            | Which.XLora
            | Which.GGUF
            | Which.GGML
            | Which.LoraGGML
            | Which.LoraGGUF
            | Which.XLoraGGML
            | Which.XLoraGGUF
            | Which.VisionPlain,
        ] = models
        self._runners: dict[str, Runner] = {}
        self._cache_lock: Lock = Lock()

    @overload
    async def chat_completion(
        self,
        *,
        config: MRSChatConfig,
        messages: list[dict[str, object]],
        stream: Literal[True],
    ) -> AsyncIterable[ChatCompletionChunkResponse]: ...

    @overload
    async def chat_completion(
        self,
        *,
        config: MRSChatConfig,
        messages: list[dict[str, object]],
    ) -> ChatCompletionResponse: ...

    async def chat_completion(
        self,
        *,
        config: MRSChatConfig,
        messages: list[dict[str, object]],
        stream: bool = False,
    ) -> AsyncIterable[ChatCompletionChunkResponse] | ChatCompletionResponse:
        if stream:
            return await self._create_chat_stream(
                messages=messages,
                model=config.model,
                temperature=config.temperature,
                top_p=config.top_p if not_missing(config.top_p) else None,
                top_k=config.top_k if not_missing(config.top_k) else None,
                max_tokens=config.max_tokens if not_missing(config.max_tokens) else None,
            )

        else:
            return await self._create_chat_completion(
                messages=messages,
                model=config.model,
                temperature=config.temperature,
                top_p=config.top_p if not_missing(config.top_p) else None,
                top_k=config.top_k if not_missing(config.top_k) else None,
                max_tokens=config.max_tokens if not_missing(config.max_tokens) else None,
            )

    async def _create_chat_completion(  # noqa: PLR0913
        self,
        model: str,
        temperature: float,
        top_p: float | None,
        top_k: int | None,
        max_tokens: int | None,
        messages: list[dict[str, object]],
    ) -> ChatCompletionResponse:
        return await self._send_chat_completion_request(
            runner=await self._get_runner(model),
            model=model,
            temperature=temperature,
            top_p=top_p,
            top_k=top_k,
            max_tokens=max_tokens,
            messages=messages,
        )

    async def _create_chat_stream(  # noqa: PLR0913
        self,
        model: str,
        temperature: float,
        top_p: float | None,
        top_k: int | None,
        max_tokens: int | None,
        messages: list[dict[str, object]],
    ) -> AsyncIterable[ChatCompletionChunkResponse]:
        return ctx.stream_sync(
            self._send_chat_completion_stream_request(
                runner=await self._get_runner(model),
                model=model,
                temperature=temperature,
                top_p=top_p,
                top_k=top_k,
                max_tokens=max_tokens,
                messages=messages,
            ),
            executor=MRS_EXECUTOR,
        )

    async def dispose(self) -> None:
        pass

    async def _get_runner(
        self,
        model_name: str,
        /,
    ) -> Runner:
        async with self._cache_lock:
            if current := self._runners.get(model_name):
                return current

            elif model := self._models.get(model_name):
                runner: Runner = await self._load_runner(model=model)
                self._runners[model_name] = runner
                return runner

            else:
                raise MRSException(
                    "Requested unsupported model - %s",
                    model_name,
                )

    @asynchronous(executor=MRS_EXECUTOR)
    def _load_runner(
        self,
        model: Which.Plain
        | Which.Lora
        | Which.XLora
        | Which.GGUF
        | Which.GGML
        | Which.LoraGGML
        | Which.LoraGGUF
        | Which.XLoraGGML
        | Which.XLoraGGUF
        | Which.VisionPlain,
    ) -> Runner:
        return Runner(which=model)

    @asynchronous(executor=MRS_EXECUTOR)
    def _send_chat_completion_request(  # noqa: PLR0913
        self,
        runner: Runner,
        model: str,
        temperature: float,
        top_p: float | None,
        top_k: int | None,
        max_tokens: int | None,
        messages: list[dict[Any, Any]],
    ) -> ChatCompletionResponse:
        return cast(
            ChatCompletionResponse,
            runner.send_chat_completion_request(
                request=ChatCompletionRequest(
                    messages=messages,
                    model=model,
                    logit_bias=None,
                    logprobs=False,
                    top_logprobs=None,
                    max_tokens=max_tokens,
                    n_choices=1,
                    presence_penalty=None,
                    frequency_penalty=None,
                    stop_seqs=None,
                    temperature=temperature,
                    top_p=top_p,
                    stream=False,
                    top_k=top_k,
                    grammar=None,
                    grammar_type=None,
                    adapters=None,
                ),
            ),
        )

    def _send_chat_completion_stream_request(  # noqa: PLR0913
        self,
        runner: Runner,
        model: str,
        temperature: float,
        top_p: float | None,
        top_k: int | None,
        max_tokens: int | None,
        messages: list[dict[Any, Any]],
    ) -> Generator[ChatCompletionChunkResponse]:
        yield from cast(
            Iterator[ChatCompletionChunkResponse],
            runner.send_chat_completion_request(
                request=ChatCompletionRequest(
                    messages=messages,
                    model=model,
                    logit_bias=None,
                    logprobs=False,
                    top_logprobs=None,
                    max_tokens=max_tokens,
                    n_choices=1,
                    presence_penalty=None,
                    frequency_penalty=None,
                    stop_seqs=None,
                    temperature=temperature,
                    top_p=top_p,
                    stream=True,
                    top_k=top_k,
                    grammar=None,
                    grammar_type=None,
                    adapters=None,
                ),
            ),
        )



from asyncio import CancelledError, iscoroutinefunction, sleep
from collections.abc import Callable, Coroutine
from typing import cast, overload

from draive.scope import ctx
from draive.types import RateLimitError
from draive.utils import mimic_function

__all__ = [
    "auto_retry",
]


@overload
def auto_retry[**Args, Result](
    function: Callable[Args, Result],
    /,
) -> Callable[Args, Result]: ...


@overload
def auto_retry[**Args, Result](
    *,
    limit: int = 1,
    delay: Callable[[int], float] | float | None = None,
) -> Callable[[Callable[Args, Result]], Callable[Args, Result]]: ...


def auto_retry[**Args, Result](
    function: Callable[Args, Result] | None = None,
    *,
    limit: int = 1,
    delay: Callable[[int], float] | float | None = None,
) -> Callable[[Callable[Args, Result]], Callable[Args, Result]] | Callable[Args, Result]:
    """\
    Simple on fail retry function wrapper. \
    Works for both sync and async functions. \
    It is not allowed to be used on class methods. \
    This wrapper is not thread safe.

    Parameters
    ----------
    function: Callable[_Args_T, _Result_T]
        function to wrap in auto retry, either sync or async
    limit: int
        limit of retries, default is 1
    delay: Callable[[], float] | float | None
        retry delay time in seconds, either concrete value or a function producing it, \
        default is None (no delay)

    Returns
    -------
    Callable[[Callable[_Args_T, _Result_T]], Callable[_Args_T, _Result_T]]
    | Callable[_Args_T, _Result_T]
        provided function wrapped in auto retry
    """

    def _wrap(
        function: Callable[Args, Result],
        /,
    ) -> Callable[Args, Result]:
        if iscoroutinefunction(function):
            return cast(
                Callable[Args, Result],
                _wrap_async(
                    function,
                    limit=limit,
                    delay=delay,
                ),
            )
        else:
            return _wrap_sync(
                function,
                limit=limit,
                delay=delay,
            )

    if function := function:
        return _wrap(function)
    else:
        return _wrap


def _wrap_sync[**Args, Result](
    function: Callable[Args, Result],
    *,
    limit: int,
    delay: Callable[[int], float] | float | None,
) -> Callable[Args, Result]:
    assert limit > 0, "Limit has to be greater than zero"  # nosec: B101
    assert delay is None, "Delay is not supported in sync wrapper"  # nosec: B101

    @mimic_function(function)
    def wrapped(
        *args: Args.args,
        **kwargs: Args.kwargs,
    ) -> Result:
        attempt: int = 0
        while True:
            try:
                return function(*args, **kwargs)
            except CancelledError as exc:
                raise exc

            except Exception as exc:
                if attempt < limit:
                    attempt += 1
                    ctx.log_error(
                        "Attempting to retry %s which failed due to an error: %s",
                        function.__name__,
                        exc,
                    )

                else:
                    raise exc

    return wrapped


def _wrap_async[**Args, Result](
    function: Callable[Args, Coroutine[None, None, Result]],
    *,
    limit: int,
    delay: Callable[[int], float] | float | None,
) -> Callable[Args, Coroutine[None, None, Result]]:
    assert limit > 0, "Limit has to be greater than zero"  # nosec: B101

    @mimic_function(function)
    async def wrapped(
        *args: Args.args,
        **kwargs: Args.kwargs,
    ) -> Result:
        attempt: int = 0
        while True:
            try:
                return await function(*args, **kwargs)
            except CancelledError as exc:
                raise exc

            except RateLimitError as exc:
                attempt += 1
                ctx.log_warning(
                    "Attempting to retry %s after %.2fs which failed due to rate limit",
                    function.__name__,
                    exc.retry_after,
                )
                await sleep(exc.retry_after)

            except Exception as exc:
                if attempt < limit:
                    attempt += 1
                    ctx.log_error(
                        "Attempting to retry %s which failed due to an error",
                        function.__name__,
                        exception=exc,
                    )

                    match delay:
                        case None:
                            continue

                        case float(strict):
                            await sleep(delay=strict)

                        case make_delay:  # type: Callable[[], float]
                            await sleep(delay=make_delay(attempt))  # pyright: ignore[reportCallIssue, reportUnknownArgumentType]

                else:
                    raise exc

    return wrapped



from collections.abc import Iterable, Sequence
from typing import Any, Literal, Protocol, runtime_checkable

from draive.instructions import Instruction
from draive.lmm import AnyTool, Toolbox
from draive.parameters import DataModel
from draive.types import MultimodalContent, MultimodalContentConvertible

__all__ = [
    "ModelGenerator",
    "ModelGeneratorDecoder",
]


@runtime_checkable
class ModelGeneratorDecoder(Protocol):
    def __call__(
        self,
        generated: MultimodalContent,
    ) -> dict[str, Any]: ...


@runtime_checkable
class ModelGenerator(Protocol):
    async def __call__[Generated: DataModel](  # noqa: PLR0913
        self,
        generated: type[Generated],
        /,
        *,
        instruction: Instruction | str,
        input: MultimodalContent | MultimodalContentConvertible,  # noqa: A002
        schema_injection: Literal["auto", "full", "simplified", "skip"] = "auto",
        tools: Toolbox | Sequence[AnyTool] | None = None,
        examples: Iterable[tuple[MultimodalContent | MultimodalContentConvertible, Generated]]
        | None = None,
        decoder: ModelGeneratorDecoder | None = None,
        **extra: Any,
    ) -> Generated: ...



import sys
from collections.abc import Callable
from inspect import Parameter, signature
from inspect import _empty as INSPECT_EMPTY  # pyright: ignore[reportPrivateUsage]
from typing import Any, Self, cast, final, overload

from draive.parameters.annotations import ParameterDefaultFactory, allows_missing
from draive.parameters.specification import ParameterSpecification, parameter_specification
from draive.parameters.validation import (
    ParameterValidationContext,
    ParameterValidationError,
    ParameterValidator,
    ParameterVerifier,
    parameter_validator,
)
from draive.utils import MISSING, Missing, freeze, is_missing, mimic_function, not_missing

__all__ = [
    "Argument",
    "ParametrizedFunction",
]


@overload
def Argument[Value](
    *,
    aliased: str | None = None,
    description: str | None = None,
    default: Value | Missing = MISSING,
    validator: ParameterValidator[Value] | Missing = MISSING,
    specification: ParameterSpecification | Missing = MISSING,
) -> Value: ...


@overload
def Argument[Value](
    *,
    aliased: str | None = None,
    description: str | None = None,
    default_factory: ParameterDefaultFactory[Value] | Missing = MISSING,
    validator: ParameterValidator[Value] | Missing = MISSING,
    specification: ParameterSpecification | Missing = MISSING,
) -> Value: ...


@overload
def Argument[Value](
    *,
    aliased: str | None = None,
    description: str | None = None,
    default: Value | Missing = MISSING,
    verifier: ParameterVerifier[Value] | None = None,
    specification: ParameterSpecification | Missing = MISSING,
) -> Value: ...


@overload
def Argument[Value](
    *,
    aliased: str | None = None,
    description: str | None = None,
    default_factory: ParameterDefaultFactory[Value] | Missing = MISSING,
    verifier: ParameterVerifier[Value] | None = None,
    specification: ParameterSpecification | Missing = MISSING,
) -> Value: ...


def Argument[Value](  # noqa: PLR0913 # Ruff - noqa: B008
    *,
    aliased: str | None = None,
    description: str | None = None,
    default: Value | Missing = MISSING,
    default_factory: ParameterDefaultFactory[Value] | Missing = MISSING,
    validator: ParameterValidator[Value] | Missing = MISSING,
    verifier: ParameterVerifier[Value] | None = None,
    specification: ParameterSpecification | Missing = MISSING,
) -> Value:  # it is actually a FunctionArgument, but type checker has to be fooled
    assert (  # nosec: B101
        is_missing(default_factory) or is_missing(default)
    ), "Can't specify both default value and factory"
    assert (  # nosec: B101
        is_missing(validator) or verifier is None
    ), "Can't specify both validator and verifier"

    return cast(
        Value,
        FunctionArgument(
            aliased=aliased,
            description=description,
            default=default,
            default_factory=default_factory,
            validator=validator,
            verifier=verifier,
            specification=specification,
        ),
    )


@final
class FunctionArgument:
    def __init__(  # noqa: PLR0913
        self,
        aliased: str | None,
        description: str | None,
        default: Any | Missing,
        default_factory: ParameterDefaultFactory[Any] | Missing,
        validator: ParameterValidator[Any] | Missing,
        verifier: ParameterVerifier[Any] | None,
        specification: ParameterSpecification | Missing,
    ) -> None:
        self.aliased: str | None = aliased
        self.description: str | None = description
        self.default: Any | Missing = default
        self.default_factory: Callable[[], Any] | Missing = default_factory
        self.validator: ParameterValidator[Any] | Missing = validator
        self.verifier: ParameterVerifier[Any] | None = verifier
        self.specification: ParameterSpecification | Missing = specification


class ParametrizedFunction[**Args, Result]:
    def __init__(
        self,
        function: Callable[Args, Result],
    ) -> None:
        assert (  # nosec: B101
            not isinstance(function, ParametrizedFunction)
        ), "Cannot parametrize the same function more than once!"

        self._call: Callable[Args, Result] = function
        globalns: dict[str, Any]
        if hasattr(function, "__globals__"):
            globalns = function.__globals__  # pyright: ignore[reportUnknownMemberType, reportAttributeAccessIssue, reportUnknownVariableType]

        else:
            globalns = sys.modules.get(function.__module__).__dict__

        self._parameters: dict[str, FunctionParameter] = {
            parameter.name: FunctionParameter.of(
                parameter,
                type_arguments={},  # TODO: verify if we should get type parameters
                globalns=globalns,  # pyright: ignore[reportUnknownArgumentType]
                localns=None,
            )
            for parameter in signature(function).parameters.values()
        }

        mimic_function(function, within=self)

    def validate_arguments(
        self,
        **arguments: Any,
    ) -> dict[str, Any]:
        # TODO: add support for positional and parametrized arguments
        context: ParameterValidationContext = ParameterValidationContext(
            path=(self.__class__.__qualname__,),
        )
        validated: dict[str, Any] = {}
        for parameter in self._parameters.values():
            validated[parameter.name] = parameter.validated(
                arguments.get(
                    parameter.name,
                    arguments.get(
                        parameter.aliased,
                        MISSING,
                    )
                    if parameter.aliased
                    else MISSING,
                ),
                context=context,
            )

        return validated

    def __call__(
        self,
        *args: Args.args,
        **kwargs: Args.kwargs,
    ) -> Result:
        assert not args, "Positional unkeyed arguments are not supported"  # nosec: B101
        return self._call(*args, **self.validate_arguments(**kwargs))  # pyright: ignore[reportCallIssue]


@final
class FunctionParameter:
    def __init__(  # noqa: PLR0913
        self,
        name: str,
        aliased: str | None,
        description: str | None,
        annotation: Any,
        default: Any | Missing,
        default_factory: ParameterDefaultFactory[Any] | Missing,
        allows_missing: bool,
        validator: ParameterValidator[Any],
        specification: ParameterSpecification | Missing,
    ) -> None:
        self.name: str = name
        self.aliased: str | None = aliased
        self.description: str | None = description
        self.annotation: Any = annotation
        self.default_value: Callable[[], Any | Missing]
        if not_missing(default_factory):
            self.default_value = default_factory

        elif not_missing(default):
            self.default_value = lambda: default

        else:
            self.default_value = lambda: MISSING

        self.has_default: bool = not_missing(default_factory) or not_missing(default)
        self.allows_missing: bool = allows_missing
        self.validator: ParameterValidator[Any] = validator
        self.specification: ParameterSpecification | Missing = specification

        freeze(self)

    def validated(
        self,
        value: Any,
        /,
        context: ParameterValidationContext,
    ) -> Any:
        if is_missing(value):
            if self.has_default:
                return self.validator(self.default_value(), context.appending_path(f"@{self.name}"))

            elif self.allows_missing:
                return MISSING

            else:
                raise ParameterValidationError.missing(
                    context=context.appending_path(f"@{self.name}")
                )

        else:
            return self.validator(value, context.appending_path(f"@{self.name}"))

    @classmethod
    def of(
        cls,
        parameter: Parameter,
        /,
        type_arguments: dict[str, Any],
        globalns: dict[str, Any],
        localns: dict[str, Any] | None,
    ) -> Self:
        if parameter.annotation is INSPECT_EMPTY:
            raise TypeError(
                "Untyped argument %s",
                parameter.name,
            )

        match parameter.default:
            case FunctionArgument() as argument:
                return cls(
                    name=parameter.name,
                    aliased=argument.aliased,
                    description=argument.description,
                    annotation=parameter.annotation,
                    default=argument.default,
                    default_factory=argument.default_factory,
                    allows_missing=allows_missing(
                        parameter.annotation,
                        type_arguments=type_arguments,
                        globalns=globalns,
                        localns=localns,
                    ),
                    validator=argument.validator
                    if not_missing(argument.validator)
                    else parameter_validator(
                        parameter.annotation,
                        verifier=argument.verifier,
                        type_arguments=type_arguments,
                        globalns=globalns,
                        localns=localns,
                        recursion_guard=frozenset(),
                    ),
                    specification=argument.specification
                    if not_missing(argument.specification)
                    else parameter_specification(
                        parameter.annotation,
                        description=argument.description,
                        type_arguments=type_arguments,
                        globalns=globalns,
                        localns=localns,
                        recursion_guard=frozenset(),
                    ),
                )

            case default:
                return cls(
                    name=parameter.name,
                    aliased=None,
                    description=None,
                    annotation=parameter.annotation,
                    default=MISSING if default is INSPECT_EMPTY else default,
                    default_factory=MISSING,
                    allows_missing=allows_missing(
                        parameter.annotation,
                        type_arguments=type_arguments,
                        globalns=globalns,
                        localns=localns,
                    ),
                    validator=parameter_validator(
                        parameter.annotation,
                        verifier=None,
                        type_arguments=type_arguments,
                        globalns=globalns,
                        localns=localns,
                        recursion_guard=frozenset(),
                    ),
                    specification=parameter_specification(
                        parameter.annotation,
                        description=None,
                        type_arguments=type_arguments,
                        globalns=globalns,
                        localns=localns,
                        recursion_guard=frozenset(),
                    ),
                )



from pathlib import Path
from typing import Any

from draive.gemini.config import GeminiConfig
from draive.scope import ctx
from draive.sentencepiece import sentencepiece_tokenize_text
from draive.utils import cache

__all__ = [
    "gemini_tokenize_text",
]


def gemini_tokenize_text(
    text: str,
    **extra: Any,
) -> list[int]:
    return sentencepiece_tokenize_text(
        text=text,
        model_path=_model_path(ctx.state(GeminiConfig).model),
        **extra,
    )


@cache(limit=2)
def _model_path(model_name: str) -> str:
    model_file: str = _mapping.get(model_name, "gemini_tokenizer.model")
    return str(Path(__file__).parent / "tokens" / model_file)


_mapping: dict[str, str] = {
    "gemini-1.5-flash": "gemini_tokenizer.model",
    "gemini-1.5-pro": "gemini_tokenizer.model",
}



from draive.anthropic.client import AnthropicClient
from draive.anthropic.config import AnthropicConfig
from draive.anthropic.errors import AnthropicException
from draive.anthropic.lmm import anthropic_lmm_invocation
from draive.anthropic.tokenization import anthropic_tokenize_text

__all__ = [
    "anthropic_lmm_invocation",
    "anthropic_tokenize_text",
    "AnthropicClient",
    "AnthropicConfig",
    "AnthropicException",
]



from typing import Literal, Required, TypedDict

from draive.parameters import ParametersSpecification

__all__ = [
    "ToolFunctionSpecification",
    "ToolSpecification",
]


class ToolFunctionSpecification(TypedDict, total=False):
    name: Required[str]
    description: Required[str]
    parameters: Required[ParametersSpecification]


class ToolSpecification(TypedDict, total=False):
    type: Required[Literal["function"]]
    function: Required[ToolFunctionSpecification]



from typing import Any, cast

from sentencepiece import SentencePieceProcessor  # pyright: ignore[reportMissingTypeStubs]

from draive.scope import ctx
from draive.sentencepiece.config import SentencePieceConfig
from draive.utils import cache, not_missing

__all__ = [
    "sentencepiece_tokenize_text",
]


def sentencepiece_tokenize_text(
    text: str,
    **extra: Any,
) -> list[int]:
    config: SentencePieceConfig = ctx.state(SentencePieceConfig).updated(**extra)
    if not_missing(config.model_path):
        return cast(
            list[int],
            _encoding(model_path=config.model_path).encode(text),  # pyright: ignore[reportUnknownMemberType, reportAttributeAccessIssue]
        )
    else:
        raise ValueError("Missing sentencepiece tokenizer model path")


@cache(limit=4)
def _encoding(model_path: str) -> SentencePieceProcessor:
    return SentencePieceProcessor(model_file=model_path)  # pyright: ignore[reportCallIssue]



from typing import Literal

from draive.parameters import DataModel
from draive.utils import MISSING, Missing

__all__ = [
    "GeminiConfig",
    "GeminiEmbeddingConfig",
]


class GeminiConfig(DataModel):
    model: str = "gemini-1.5-flash"
    temperature: float = 0.75
    response_format: Literal["text/plain", "application/json"] | Missing = MISSING
    top_p: float | Missing = MISSING
    top_k: int | Missing = MISSING
    max_tokens: int = 2048
    timeout: float | Missing = MISSING


class GeminiEmbeddingConfig(DataModel):
    model: str = "embedding-gecko-001"
    batch_size: int = 128



from os import environ, getenv
from typing import overload

__all__ = [
    "getenv_bool",
    "getenv_int",
    "getenv_float",
    "getenv_str",
    "load_env",
]


@overload
def getenv_bool(key: str) -> bool | None:
    ...


@overload
def getenv_bool(key: str, default: bool) -> bool:
    ...


def getenv_bool(
    key: str,
    default: bool | None = None,
) -> bool | None:
    if value := getenv(key=key):
        return value.lower() in ("true", "1", "t")
    else:
        return default


@overload
def getenv_int(key: str) -> int | None:
    ...


@overload
def getenv_int(key: str, default: int) -> int:
    ...


def getenv_int(
    key: str,
    default: int | None = None,
) -> int | None:
    if value := getenv(key=key):
        return int(value)

    else:
        return default


@overload
def getenv_float(key: str) -> float | None:
    ...


@overload
def getenv_float(key: str, default: float) -> float:
    ...


def getenv_float(
    key: str,
    default: float | None = None,
) -> float | None:
    if value := getenv(key=key):
        return float(value)

    else:
        return default


@overload
def getenv_str(key: str) -> str | None:
    ...


@overload
def getenv_str(key: str, default: str) -> str:
    ...


def getenv_str(
    key: str,
    default: str | None = None,
) -> str | None:
    if value := getenv(key=key):
        return value
    else:
        return default


def load_env(
    path: str | None = None,
    override: bool = True,
) -> None:
    """\
    Minimalist implementation of environment variables file loader. \
    When the file is not available configuration won't be loaded.
    Allows only subset of formatting:
    - lines starting with '#' are ignored
    - other comments are not allowed
    - each element is in a new line
    - each element must be a `key=value` pair without whitespaces or additional characters
    - keys without values are ignored

    Parameters
    ----------
    path: str
        custom path to load environment variables, default is '.env'
    override: bool
        override existing variables on conflict if True, otherwise keep existing
    """

    try:
        with open(file=path or ".env") as file:
            for line in file.readlines():
                if line.startswith("#"):
                    continue  # ignore commented

                idx: int  # find where key ends
                for element in enumerate(line):
                    if element[1] == "=":
                        idx: int = element[0]
                        break
                else:  # ignore keys without assignment
                    continue

                if idx >= len(line):
                    continue  # ignore keys without values

                key: str = line[0:idx]
                value: str = line[idx + 1 :].strip()
                if value and (override or key not in environ):
                    environ[key] = value
    except FileNotFoundError:
        pass  # ignore loading if no .env available



from draive.parameters.annotations import ParameterDefaultFactory
from draive.parameters.basic import BasicValue
from draive.parameters.data import Field, ParametrizedData
from draive.parameters.errors import ParameterValidationContext, ParameterValidationError

# from draive.parameters.definition import ParametersDefinition
from draive.parameters.function import Argument, ParametrizedFunction
from draive.parameters.model import DataModel
from draive.parameters.path import ParameterPath
from draive.parameters.requirement import ParameterRequirement
from draive.parameters.specification import ParameterSpecification, ParametersSpecification
from draive.parameters.state import State, Stateless
from draive.parameters.validation import ParameterValidator, ParameterVerifier

__all__ = [
    "Argument",
    "BasicValue",
    "DataModel",
    "Field",
    "ParameterDefaultFactory",
    "ParameterPath",
    "ParameterRequirement",
    "ParameterSpecification",
    "ParametersSpecification",
    "ParameterValidationContext",
    "ParameterValidationContext",
    "ParameterValidationError",
    "ParameterValidator",
    "ParameterVerifier",
    "ParametrizedData",
    "ParametrizedData",
    "ParametrizedFunction",
    "State",
    "Stateless",
]



from typing import Literal, TypedDict

from draive.parameters import DataModel
from draive.utils import MISSING, Missing

__all__ = [
    "OpenAIChatConfig",
    "OpenAIEmbeddingConfig",
    "OpenAIImageGenerationConfig",
    "OpenAISystemFingerprint",
]


class ResponseFormat(TypedDict):
    type: Literal["text", "json_object"]


class OpenAIChatConfig(DataModel):
    model: str = "gpt-4o-mini"
    temperature: float = 0.75
    top_p: float | Missing = MISSING
    frequency_penalty: float | Missing = MISSING
    max_tokens: int = 2048
    seed: int | None | Missing = MISSING
    response_format: ResponseFormat | Missing = MISSING
    vision_details: Literal["auto", "low", "high"] | Missing = MISSING
    timeout: float | Missing = MISSING


class OpenAIEmbeddingConfig(DataModel):
    model: str = "text-embedding-3-small"
    dimensions: int | Missing = MISSING
    batch_size: int = 128
    encoding_format: Literal["float", "base64"] | Missing = MISSING
    timeout: float | Missing = MISSING


class OpenAIImageGenerationConfig(DataModel):
    model: str = "dall-e-2"
    quality: Literal["standard", "hd"] = "standard"
    size: Literal["256x256", "512x512", "1024x1024", "1792x1024", "1024x1792"] = "1024x1024"
    style: Literal["vivid", "natural"] = "vivid"
    timeout: float | Missing = MISSING
    response_format: Literal["url", "b64_json"] = "b64_json"


class OpenAISystemFingerprint(DataModel):
    system_fingerprint: str



from draive.parameters import DataModel
from draive.utils import MISSING, Missing

__all__ = [
    "SentencePieceConfig",
]


class SentencePieceConfig(DataModel):
    model_path: str | Missing = MISSING
